/***************************************************************************

						cave tilemap manager

					 ( customized "tilemap.c")

***************************************************************************/
#ifndef DECLARE

#include "state.h"

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


#define CAVE_ALL_TILEMAPS	0
#define SWAP(X,Y) { int temp = X; X = Y; Y = temp; }

static struct cave_tile_info
{
	const UINT8 *pen_data;
	const UINT32 *pal_data;
	UINT32 tile_number;
	UINT32 pen_usage;
	UINT32 flags;	/* flipx, flipy, ignore_transparency */
	UINT32 priority;
} cave_tile_info;

#define CAVE_SET_TILE_INFO(GFX,CODE,COLOR,FLAGS) { \
	const struct GfxElement *gfx = Machine->gfx[(GFX)]; \
	int _code = (CODE) % gfx->total_elements; \
	cave_tile_info.tile_number = _code; \
	cave_tile_info.pen_data = gfx->gfxdata + _code*gfx->char_modulo; \
	cave_tile_info.pal_data = &gfx->colortable[gfx->color_granularity * (COLOR)]; \
	cave_tile_info.pen_usage = gfx->pen_usage?gfx->pen_usage[_code]:0; \
	cave_tile_info.flags = FLAGS; \
	if (gfx->flags & GFX_PACKED) cave_tile_info.flags |= CAVE_TILE_4BPP; \
	if (gfx->flags & GFX_SWAPXY) cave_tile_info.flags |= CAVE_TILE_SWAPXY; \
}

#define CAVE_TILE_FLIPX					0x01
#define CAVE_TILE_FLIPY					0x02
#define CAVE_TILE_SWAPXY					0x04
#define CAVE_TILE_4BPP					0x10



static int cave_tilemap_init( void );
static void cave_tilemap_close( void );
static struct cave_tilemap *cave_tilemap_create(
	void (*tile_get_info)( int memory_offset ),
	int tile_width, int tile_height, /* in pixels */
	int num_cols, int num_rows /* in tiles */
);

static void cave_tilemap_dispose( struct cave_tilemap *tilemap );

static void cave_tilemap_set_transparent_pen( struct cave_tilemap *tilemap, int pen );

static void cave_tilemap_mark_tile_dirty( struct cave_tilemap *tilemap, int memory_offset );
static void cave_tilemap_mark_all_tiles_dirty( struct cave_tilemap *tilemap );

static void cave_tilemap_set_scroll_rows( struct cave_tilemap *tilemap, int scroll_rows ); /* default: 1 */
static void cave_tilemap_set_scrolldx( struct cave_tilemap *tilemap, int dx, int dx_if_flipped );
static void cave_tilemap_set_scrollx( struct cave_tilemap *tilemap, int row, int value );

static void cave_tilemap_set_scroll_cols( struct cave_tilemap *tilemap, int scroll_cols ); /* default: 1 */
static void cave_tilemap_set_scrolldy( struct cave_tilemap *tilemap, int dy, int dy_if_flipped );
static void cave_tilemap_set_scrolly( struct cave_tilemap *tilemap, int col, int value );

#define CAVE_TILEMAP_FLIPX 0x1
#define CAVE_TILEMAP_FLIPY 0x2
static void cave_tilemap_set_flip( struct cave_tilemap *tilemap, int attributes );
static void cave_tilemap_set_clip( struct cave_tilemap *tilemap, const struct rectangle *clip );
static void cave_tilemap_set_enable( struct cave_tilemap *tilemap, int enable );

static void cave_tilemap_update( struct cave_tilemap *tilemap );
static void cave_tilemap_draw( struct osd_bitmap *dest, struct cave_tilemap *tilemap, UINT32 flags );


static UINT32 cave_tilemap_scan_rows( UINT32 col, UINT32 row, UINT32 num_cols, UINT32 num_rows );

static void tilemap_mark_all_pixels_dirty( struct cave_tilemap *tilemap );

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

//#define MAX_TILESIZE 32
#define MAX_TILESIZE 16

struct cached_tile_info
{
	const UINT8 *pen_data;
	const UINT32 *pal_data;
	UINT32 pen_usage;
	UINT32 flags;
};

struct cave_tilemap_mask
{
	struct osd_bitmap *bitmask;
	int line_offset;
	UINT8 *data;
	UINT8 **data_row;
};

struct cave_tilemap
{
	int *memory_offset_to_cached_indx;
	UINT32 *cached_indx_to_memory_offset;
	int logical_flip_to_cached_flip[4];

	/* callback to interpret video VRAM for the tilemap */
	void (*tile_get_info)( int memory_offset );

	UINT32 max_memory_offset;
	UINT32 num_tiles;
	UINT32 num_logical_rows, num_logical_cols;
	UINT32 num_cached_rows, num_cached_cols;
	UINT32 tile_size;
	UINT32 num_pens;
	UINT32 cached_width, cached_height; /* size in pixels of tilemap as a whole */

	struct cached_tile_info *cached_tile_info;

	int dx, dx_if_flipped;
	int dy, dy_if_flipped;
	int scrollx_delta, scrolly_delta;

	int enable;
	int attributes;

	int transparent_pen;
	UINT32 fgmask[4];

	int bNeedRender;

	UINT32 *pPenToPixel[8];

	void (*draw_tile)( struct cave_tilemap *tilemap, UINT32 cached_indx, UINT32 col, UINT32 row );

	void (*draw)( int, int );

	UINT8 *priority,	/* priority for each tile */
		**priority_row;

	UINT8 *visible; /* boolean flag for each tile */

	UINT8 *dirty_vram; /* boolean flag for each tile */

	UINT8 *dirty_pixels;

	int scroll_rows, scroll_cols;
	int *rowscroll, *colscroll;

	int orientation;
	int clip_left,clip_right,clip_top,clip_bottom;

	/* cached color data */
	struct osd_bitmap *pixmap;
	int pixmap_line_offset;

	struct cave_tilemap_mask *foreground;

	struct cave_tilemap *next; /* resource tracking */
};

static UINT8 flip_bit_table[0x100]; /* horizontal flip for 8 pixels */
static struct cave_tilemap *first_tilemap; /* resource tracking */
static int screen_width, screen_height;

enum
{
	TILE_TRANSPARENT,
	TILE_MASKED,
	TILE_OPAQUE
};

static struct
{
	int clip_left, clip_top, clip_right, clip_bottom;
	int source_width, source_height;
	int dest_line_offset,source_line_offset,mask_line_offset;
	int dest_row_offset,source_row_offset,mask_row_offset;
	struct osd_bitmap *screen, *pixmap, *bitmask;
	UINT8 **mask_data_row;
	UINT8 **priority_data_row;
	int tile_priority;
} tmblt;

static int PenToPixel_Init( struct cave_tilemap *tilemap )
{
	/*
		Construct a table for all tile orientations in advance.
		This simplifies drawing tiles and masks tremendously.
		If performance is an issue, we can always (re)introduce
		customized code for each case and forgo tables.
	*/
	int i,x,y,tx,ty;
	int tile_size = tilemap->tile_size;
	UINT32 *pPenToPixel;
	int lError;

	lError = 0;
	for( i=0; i<8; i++ )
	{
		pPenToPixel = malloc( tilemap->num_pens*sizeof(UINT32) );
		if( pPenToPixel==NULL )
		{
			lError = 1;
		}
		else
		{
			tilemap->pPenToPixel[i] = pPenToPixel;
			for( ty=0; ty<tile_size; ty++ )
			{
				for( tx=0; tx<tile_size; tx++ )
				{
					if( i&CAVE_TILE_SWAPXY )
					{
						x = ty;
						y = tx;
					}
					else
					{
						x = tx;
						y = ty;
					}
					if( i&CAVE_TILE_FLIPX ) x = tile_size-1-x;
					if( i&CAVE_TILE_FLIPY ) y = tile_size-1-y;
					*pPenToPixel++ = x+y*MAX_TILESIZE;
				}
			}
		}
	}
	return lError;
}

static void PenToPixel_Term( struct cave_tilemap *tilemap )
{
	int i;
	for( i=0; i<8; i++ )
	{
		free( tilemap->pPenToPixel[i] );
	}
}

static void tmap_render( struct cave_tilemap *tilemap )
{
	if( tilemap->bNeedRender ){
		tilemap->bNeedRender = 0;
		if( tilemap->enable ){
			UINT8 *dirty_pixels = tilemap->dirty_pixels;
			const UINT8 *visible = tilemap->visible;
			UINT32 cached_indx = 0;
			UINT32 row,col;

			/* walk over cached rows/cols (better to walk screen coords) */
			for( row=0; row<tilemap->num_cached_rows; row++ ){
				for( col=0; col<tilemap->num_cached_cols; col++ ){
					if( visible[cached_indx] && dirty_pixels[cached_indx] ){
						tilemap->draw_tile( tilemap, cached_indx, col, row );
						dirty_pixels[cached_indx] = 0;
					}
					cached_indx++;
				} /* next col */
			} /* next row */
		}
	}
}



static void cave_tilemap_set_transparent_pen( struct cave_tilemap *tilemap, int pen )
{
	tilemap->transparent_pen = pen;
}


/***********************************************************************************/
static UINT32 cave_tilemap_scan_rows( UINT32 col, UINT32 row, UINT32 num_cols, UINT32 num_rows )
{
	/* logical (col,row) -> memory offset */
	return row*num_cols + col;
}
/*********************************************************************************/

static struct osd_bitmap *create_tmpbitmap( int width, int height, int depth )
{
	return osd_alloc_bitmap( width,height,depth );
}

static struct osd_bitmap *create_bitmask( int width, int height )
{
	width = (width+7)/8; /* 8 bits per byte */
	return osd_alloc_bitmap( width,height, 8 );
}

/***********************************************************************************/

static int mappings_create( struct cave_tilemap *tilemap )
{
	int max_memory_offset = 0;
	UINT32 col,row;
	UINT32 num_logical_rows = tilemap->num_logical_rows;
	UINT32 num_logical_cols = tilemap->num_logical_cols;
	/* count offsets (might be larger than num_tiles) */
	for( row=0; row<num_logical_rows; row++ )
	{
		for( col=0; col<num_logical_cols; col++ )
		{
			UINT32 memory_offset = cave_tilemap_scan_rows( col, row, num_logical_cols, num_logical_rows );
			if( memory_offset>max_memory_offset ) max_memory_offset = memory_offset;
		}
	}
	max_memory_offset++;
	tilemap->max_memory_offset = max_memory_offset;
	/* logical to cached (tilemap_mark_dirty) */
	tilemap->memory_offset_to_cached_indx = malloc( sizeof(int)*max_memory_offset );
	if( tilemap->memory_offset_to_cached_indx )
	{
		/* cached to logical (get_tile_info) */
		tilemap->cached_indx_to_memory_offset = malloc( sizeof(UINT32)*tilemap->num_tiles );
		if( tilemap->cached_indx_to_memory_offset ) return 0; /* no error */
		free( tilemap->memory_offset_to_cached_indx );
	}
	return -1; /* error */
}

static void mappings_dispose( struct cave_tilemap *tilemap )
{
	free( tilemap->cached_indx_to_memory_offset );
	free( tilemap->memory_offset_to_cached_indx );
}

static void mappings_update( struct cave_tilemap *tilemap )
{
	int logical_flip;
	UINT32 logical_indx, cached_indx;
	UINT32 num_cached_rows = tilemap->num_cached_rows;
	UINT32 num_cached_cols = tilemap->num_cached_cols;
	UINT32 num_logical_rows = tilemap->num_logical_rows;
	UINT32 num_logical_cols = tilemap->num_logical_cols;
	for( logical_indx=0; logical_indx<tilemap->max_memory_offset; logical_indx++ )
	{
		tilemap->memory_offset_to_cached_indx[logical_indx] = -1;
	}

	for( logical_indx=0; logical_indx<tilemap->num_tiles; logical_indx++ )
	{
		UINT32 logical_col = logical_indx%num_logical_cols;
		UINT32 logical_row = logical_indx/num_logical_cols;
		int memory_offset = cave_tilemap_scan_rows( logical_col, logical_row, num_logical_cols, num_logical_rows );
		UINT32 cached_col = logical_col;
		UINT32 cached_row = logical_row;
		if( tilemap->orientation & ORIENTATION_SWAP_XY ) SWAP(cached_col,cached_row)
		if( tilemap->orientation & ORIENTATION_FLIP_X ) cached_col = (num_cached_cols-1)-cached_col;
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) cached_row = (num_cached_rows-1)-cached_row;
		cached_indx = cached_row*num_cached_cols+cached_col;
		tilemap->memory_offset_to_cached_indx[memory_offset] = cached_indx;
		tilemap->cached_indx_to_memory_offset[cached_indx] = memory_offset;
	}
	for( logical_flip = 0; logical_flip<4; logical_flip++ )
	{
		int cached_flip = logical_flip;
		if( tilemap->attributes&CAVE_TILEMAP_FLIPX ) cached_flip ^= CAVE_TILE_FLIPX;
		if( tilemap->attributes&CAVE_TILEMAP_FLIPY ) cached_flip ^= CAVE_TILE_FLIPY;
#ifndef PREROTATE_GFX
		if( Machine->orientation & ORIENTATION_SWAP_XY )
		{
			if( Machine->orientation & ORIENTATION_FLIP_X ) cached_flip ^= CAVE_TILE_FLIPY;
			if( Machine->orientation & ORIENTATION_FLIP_Y ) cached_flip ^= CAVE_TILE_FLIPX;
		}
		else
		{
			if( Machine->orientation & ORIENTATION_FLIP_X ) cached_flip ^= CAVE_TILE_FLIPX;
			if( Machine->orientation & ORIENTATION_FLIP_Y ) cached_flip ^= CAVE_TILE_FLIPY;
		}
#endif
		if( tilemap->orientation & ORIENTATION_SWAP_XY )
		{
			cached_flip = ((cached_flip&1)<<1) | ((cached_flip&2)>>1);
		}
		tilemap->logical_flip_to_cached_flip[logical_flip] = cached_flip;
	}
}

/***********************************************************************************/

/**** DEPTH == 16 ****/

#define DEPTH 16
#define TILE_SIZE	8
#define DATA_TYPE UINT16
#define DECLARE(function,args,body) static void function##8x8x16BPP args body
#include "vidhrdw/cavetmap.c"

#define TILE_SIZE	16
#define DATA_TYPE UINT16
#define DECLARE(function,args,body) static void function##16x16x16BPP args body
#include "vidhrdw/cavetmap.c"

#undef DEPTH

/*********************************************************************************/

static void mask_dispose( struct cave_tilemap_mask *mask )
{
	if( mask )
	{
		free( mask->data_row );
		free( mask->data );
		osd_free_bitmap( mask->bitmask );
		free( mask );
	}
}

static struct cave_tilemap_mask *mask_create( struct cave_tilemap *tilemap )
{
	int row;
	struct cave_tilemap_mask *mask = malloc( sizeof(struct cave_tilemap_mask) );
	if( mask )
	{
		mask->data = malloc( tilemap->num_tiles );
		mask->data_row = malloc( tilemap->num_cached_rows * sizeof(UINT8 *) );
		mask->bitmask = create_bitmask( tilemap->cached_width, tilemap->cached_height );
		if( mask->data && mask->data_row && mask->bitmask )
		{
			for( row=0; row<tilemap->num_cached_rows; row++ )
			{
				mask->data_row[row] = mask->data + row*tilemap->num_cached_cols;
			}
			mask->line_offset = mask->bitmask->line[1] - mask->bitmask->line[0];
			return mask;
		}
	}
	mask_dispose( mask );
	return NULL;
}

/***********************************************************************************/

static void install_draw_handlers( struct cave_tilemap *tilemap )
{
	int tile_size = tilemap->tile_size;
	tilemap->draw = NULL;
	tilemap->draw_tile = draw_tile8x8x16BPP;

	if( tile_size==8 )
	{
		tilemap->draw = draw8x8x16BPP;
	}
	else if( tile_size==16 )
	{
		tilemap->draw = draw16x16x16BPP;
	}
}

/***********************************************************************************/

static void tilemap_reset(void)
{
	cave_tilemap_mark_all_tiles_dirty(CAVE_ALL_TILEMAPS);
}

static int cave_tilemap_init( void )
{
	UINT32 value, data, bit;
	for( value=0; value<0x100; value++ )
	{
		data = 0;
		for( bit=0; bit<8; bit++ ) if( (value>>bit)&1 ) data |= 0x80>>bit;
		flip_bit_table[value] = data;
	}
	screen_width = Machine->scrbitmap->width;
	screen_height = Machine->scrbitmap->height;
	first_tilemap = 0;
	state_save_register_func_postload(tilemap_reset);			//kschk
	return 0;
}

static void cave_tilemap_close( void )
{
	while( first_tilemap )
	{
		struct cave_tilemap *next = first_tilemap->next;
		cave_tilemap_dispose( first_tilemap );
		first_tilemap = next;
	}
}

/***********************************************************************************/

static struct cave_tilemap *cave_tilemap_create(
	void (*tile_get_info)( int memory_offset ),
	int tile_width, int tile_height,
	int num_cols, int num_rows )
{
	struct cave_tilemap *tilemap;

	if( tile_width != tile_height )
	{
		logerror( "cave_tilemap_create: tile_width must be equal to tile_height\n" );
		return 0;
	}

	tilemap = calloc( 1,sizeof( struct cave_tilemap ) );
	if( tilemap )
	{
		int num_tiles = num_cols*num_rows;
		tilemap->num_logical_cols = num_cols;
		tilemap->num_logical_rows = num_rows;
		if( Machine->orientation & ORIENTATION_SWAP_XY )
		{
			SWAP( num_cols,num_rows )
		}
		tilemap->num_cached_cols = num_cols;
		tilemap->num_cached_rows = num_rows;
		tilemap->num_tiles = num_tiles;
		tilemap->num_pens = tile_width*tile_height;
		tilemap->tile_size = tile_width; /* tile_width and tile_height are equal */
		tilemap->cached_width = tile_width*num_cols;
		tilemap->cached_height = tile_height*num_rows;
		tilemap->tile_get_info = tile_get_info;
		tilemap->orientation = Machine->orientation;

		/* various defaults */
		tilemap->enable = 1;
		tilemap->scroll_rows = 1;
		tilemap->scroll_cols = 1;
		tilemap->transparent_pen = -1;

		tilemap->cached_tile_info = calloc( num_tiles, sizeof(struct cached_tile_info) );
		tilemap->priority = calloc( num_tiles,1 );
		tilemap->visible = calloc( num_tiles,1 );
		tilemap->dirty_vram = malloc( num_tiles );
		tilemap->dirty_pixels = malloc( num_tiles );
		tilemap->rowscroll = calloc(tilemap->cached_height,sizeof(int));
		tilemap->colscroll = calloc(tilemap->cached_width,sizeof(int));
		tilemap->priority_row = malloc( sizeof(UINT8 *)*num_rows );
		tilemap->pixmap = create_tmpbitmap( tilemap->cached_width, tilemap->cached_height, Machine->scrbitmap->depth );
		tilemap->foreground = mask_create( tilemap );

		if( tilemap->cached_tile_info &&
			tilemap->priority && tilemap->visible &&
			tilemap->dirty_vram && tilemap->dirty_pixels &&
			tilemap->rowscroll && tilemap->colscroll &&
			tilemap->priority_row &&
			tilemap->pixmap && tilemap->foreground &&
			(mappings_create( tilemap )==0) )
		{
			UINT32 row;
			for( row=0; row<num_rows; row++ ){
				tilemap->priority_row[row] = tilemap->priority+num_cols*row;
			}
			install_draw_handlers( tilemap );
			mappings_update( tilemap );
			cave_tilemap_set_clip( tilemap, &Machine->visible_area );
			memset( tilemap->dirty_vram, 1, num_tiles );
			memset( tilemap->dirty_pixels, 1, num_tiles );
			tilemap->pixmap_line_offset = tilemap->pixmap->line[1] - tilemap->pixmap->line[0];
			tilemap->next = first_tilemap;
			first_tilemap = tilemap;
			if( PenToPixel_Init( tilemap ) == 0 )
			{
				return tilemap;
			}
		}
		cave_tilemap_dispose( tilemap );
	}
	return 0;
}

static void cave_tilemap_dispose( struct cave_tilemap *tilemap )
{
	struct cave_tilemap *prev;

	if( tilemap==first_tilemap )
	{
		first_tilemap = tilemap->next;
	}
	else
	{
		prev = first_tilemap;
		while( prev->next != tilemap ) prev = prev->next;
		prev->next =tilemap->next;
	}
	PenToPixel_Term( tilemap );
	free( tilemap->cached_tile_info );
	free( tilemap->priority );
	free( tilemap->visible );
	free( tilemap->dirty_vram );
	free( tilemap->dirty_pixels );
	free( tilemap->rowscroll );
	free( tilemap->colscroll );
	free( tilemap->priority_row );
	osd_free_bitmap( tilemap->pixmap );
	mask_dispose( tilemap->foreground );
	mappings_dispose( tilemap );
	free( tilemap );
}

/***********************************************************************************/

static void cave_tilemap_set_enable( struct cave_tilemap *tilemap, int enable )
{
	tilemap->enable = enable?1:0;
}

static void cave_tilemap_set_flip( struct cave_tilemap *tilemap, int attributes )
{
	if( tilemap==CAVE_ALL_TILEMAPS )
	{
		tilemap = first_tilemap;
		while( tilemap )
		{
			cave_tilemap_set_flip( tilemap, attributes );
			tilemap = tilemap->next;
		}
	}
	else if( tilemap->attributes!=attributes )
	{
		tilemap->attributes = attributes;
		tilemap->orientation = Machine->orientation;
		if( attributes&CAVE_TILEMAP_FLIPY )
		{
			tilemap->orientation ^= ORIENTATION_FLIP_Y;
			tilemap->scrolly_delta = tilemap->dy_if_flipped;
		}
		else
		{
			tilemap->scrolly_delta = tilemap->dy;
		}
		if( attributes&CAVE_TILEMAP_FLIPX )
		{
			tilemap->orientation ^= ORIENTATION_FLIP_X;
			tilemap->scrollx_delta = tilemap->dx_if_flipped;
		}
		else
		{
			tilemap->scrollx_delta = tilemap->dx;
		}

		mappings_update( tilemap );
		cave_tilemap_mark_all_tiles_dirty( tilemap );
	}
}

static void cave_tilemap_set_clip( struct cave_tilemap *tilemap, const struct rectangle *clip )
{
	int left,top,right,bottom;
	if( clip )
	{
		left = clip->min_x;
		top = clip->min_y;
		right = clip->max_x+1;
		bottom = clip->max_y+1;
		if( tilemap->orientation & ORIENTATION_SWAP_XY )
		{
			SWAP(left,top)
			SWAP(right,bottom)
		}
		if( tilemap->orientation & ORIENTATION_FLIP_X )
		{
			SWAP(left,right)
			left = screen_width-left;
			right = screen_width-right;
		}
		if( tilemap->orientation & ORIENTATION_FLIP_Y )
		{
			SWAP(top,bottom)
			top = screen_height-top;
			bottom = screen_height-bottom;
		}
	}
	else
	{
		left = 0;
		top = 0;
		right = tilemap->cached_width;
		bottom = tilemap->cached_height;
	}
	tilemap->clip_left = left;
	tilemap->clip_right = right;
	tilemap->clip_top = top;
	tilemap->clip_bottom = bottom;
}

/***********************************************************************************/

static void cave_tilemap_set_scroll_cols( struct cave_tilemap *tilemap, int n )
{
	if( tilemap->orientation & ORIENTATION_SWAP_XY )
	{
		if (tilemap->scroll_rows != n)
		{
			tilemap->scroll_rows = n;
		}
	}
	else
	{
		if (tilemap->scroll_cols != n)
		{
			tilemap->scroll_cols = n;
		}
	}
}

static void cave_tilemap_set_scroll_rows( struct cave_tilemap *tilemap, int n )
{
	if( tilemap->orientation & ORIENTATION_SWAP_XY )
	{
		if (tilemap->scroll_cols != n)
		{
			tilemap->scroll_cols = n;
		}
	}
	else
	{
		if (tilemap->scroll_rows != n)
		{
			tilemap->scroll_rows = n;
		}
	}
}

/***********************************************************************************/

static void cave_tilemap_mark_tile_dirty( struct cave_tilemap *tilemap, int memory_offset )
{
	if( memory_offset<tilemap->max_memory_offset )
	{
		int cached_indx = tilemap->memory_offset_to_cached_indx[memory_offset];
		if( cached_indx>=0 )
		{
			tilemap->dirty_vram[cached_indx] = 1;
		}
	}
}

static void cave_tilemap_mark_all_tiles_dirty( struct cave_tilemap *tilemap )
{
	if( tilemap==CAVE_ALL_TILEMAPS )
	{
		tilemap = first_tilemap;
		while( tilemap )
		{
			cave_tilemap_mark_all_tiles_dirty( tilemap );
			tilemap = tilemap->next;
		}
	}
	else
	{
		memset( tilemap->dirty_vram, 1, tilemap->num_tiles );
	}
}

static void tilemap_mark_all_pixels_dirty( struct cave_tilemap *tilemap )
{
	if( tilemap==CAVE_ALL_TILEMAPS )
	{
		tilemap = first_tilemap;
		while( tilemap )
		{
			tilemap_mark_all_pixels_dirty( tilemap );
			tilemap = tilemap->next;
		}
	}
	else
	{
		/* invalidate all offscreen tiles */
		UINT32 cached_tile_indx;
		for( cached_tile_indx=0; cached_tile_indx<tilemap->num_tiles; cached_tile_indx++ )
		{
			if( !tilemap->visible[cached_tile_indx] )
			{
				tilemap->dirty_vram[cached_tile_indx] = 1;
			}
		}
		memset( tilemap->dirty_pixels, 1, tilemap->num_tiles );
	}
}


/***********************************************************************************/

static void draw_pen_mask(
	struct cave_tilemap *tilemap,
	struct osd_bitmap *mask,
	UINT32 x0, UINT32 y0,
	UINT32 tile_size,
	const UINT8 *pPenData,
	int transparent_pen,
	UINT32 flags,
	int pitch )
{
	UINT32 *pPenToPixel = tilemap->pPenToPixel[flags&(CAVE_TILE_SWAPXY|CAVE_TILE_FLIPY|CAVE_TILE_FLIPX)];
	int tx,ty;
	const UINT8 *pSource;
	UINT8 data;
	UINT32 yx;

	if( flags&CAVE_TILE_4BPP )
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size/2; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				if( (data&0xf)!=transparent_pen )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
				yx = *pPenToPixel++;
				if( (data>>4)!=transparent_pen )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
			}
			pPenData += pitch/2;
		}
	}
	else
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				if( data!=transparent_pen )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
			}
			pPenData += pitch;
		}
	}
}

static void draw_mask(
	struct cave_tilemap *tilemap,
	struct osd_bitmap *mask,
	UINT32 x0, UINT32 y0,
	UINT32 tile_size,
	const UINT8 *pPenData,
	UINT32 transmask,
	UINT32 flags,
	int pitch )
{
	UINT32 *pPenToPixel = tilemap->pPenToPixel[flags&(CAVE_TILE_SWAPXY|CAVE_TILE_FLIPY|CAVE_TILE_FLIPX)];
	int tx,ty;
	const UINT8 *pSource;
	UINT8 data;
	UINT32 yx;

	if( flags&CAVE_TILE_4BPP )
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size/2; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				if( !((1<<(data&0xf))&transmask) )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
				yx = *pPenToPixel++;
				if( !((1<<(data>>4))&transmask) )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
			}
			pPenData += pitch/2;
		}
	}
	else
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				if( !((1<<data)&transmask) )
				{
					mask->line[y0+yx/MAX_TILESIZE][(x0+(yx%MAX_TILESIZE))/8] |= 0x80>>(yx%8);
				}
			}
			pPenData += pitch;
		}
	}
}

static void ClearMask( struct osd_bitmap *bitmap, int tile_size, int x0, int y0 )
{
	UINT8 *pDest;
	int ty,tx;
	for( ty=0; ty<tile_size; ty++ )
	{
		pDest = bitmap->line[y0+ty]+x0/8;
		for( tx=tile_size/8; tx!=0; tx-- )
		{
			*pDest++ = 0x00;
		}
	}
}

static int InspectMask( struct osd_bitmap *bitmap, int tile_size, int x0, int y0 )
{
	const UINT8 *pSource;
	int ty,tx;

	switch( bitmap->line[y0][x0/8] )
	{
	case 0xff: /* possibly opaque */
		for( ty=0; ty<tile_size; ty++ )
		{
			pSource = bitmap->line[y0+ty]+x0/8;
			for( tx=tile_size/8; tx!=0; tx-- )
			{
				if( *pSource++ != 0xff ) return TILE_MASKED;
			}
		}
		return TILE_OPAQUE;

	case 0x00: /* possibly transparent */
		for( ty=0; ty<tile_size; ty++ )
		{
			pSource = bitmap->line[y0+ty]+x0/8;
			for( tx=tile_size/8; tx!=0; tx-- )
			{
				if( *pSource++ != 0x00 ) return TILE_MASKED;
			}
		}
		return TILE_TRANSPARENT;

	default:
		return TILE_MASKED;
	}
}

static void render_mask( struct cave_tilemap *tilemap, UINT32 cached_indx )
{
	const struct cached_tile_info *cached_tile_info = &tilemap->cached_tile_info[cached_indx];
	UINT32 col = cached_indx%tilemap->num_cached_cols;
	UINT32 row = cached_indx/tilemap->num_cached_cols;
	UINT32 tile_size = tilemap->tile_size;
	UINT32 y0 = tile_size*row;
	UINT32 x0 = tile_size*col;
	int pitch = tile_size;
	UINT32 pen_usage = cached_tile_info->pen_usage;
	const UINT8 *pen_data = cached_tile_info->pen_data;
	UINT32 flags = cached_tile_info->flags;

	if( pen_usage )
	{
		UINT32 fgmask = 1 << tilemap->transparent_pen;
	 	if( flags&TILE_IGNORE_TRANSPARENCY ) fgmask = 0;
		if( pen_usage == fgmask )
		{
			tilemap->foreground->data_row[row][col] = TILE_TRANSPARENT;
		}
		else if( pen_usage & fgmask )
		{
			ClearMask( tilemap->foreground->bitmask, tile_size, x0, y0 );
			draw_mask( tilemap,tilemap->foreground->bitmask,
				x0, y0, tile_size, pen_data, fgmask, flags, pitch );
			tilemap->foreground->data_row[row][col] = TILE_MASKED;
		}
		else
		{
			tilemap->foreground->data_row[row][col] = TILE_OPAQUE;
		}
	}
	else
	{
		ClearMask( tilemap->foreground->bitmask, tile_size, x0, y0 );
		draw_pen_mask(
				tilemap,tilemap->foreground->bitmask,
				x0, y0, tile_size, pen_data, tilemap->transparent_pen, flags, pitch );
		tilemap->foreground->data_row[row][col] =
			InspectMask( tilemap->foreground->bitmask, tile_size, x0, y0 );
	}
}

static void update_tile_info( struct cave_tilemap *tilemap )
{
	int *logical_flip_to_cached_flip = tilemap->logical_flip_to_cached_flip;
	UINT32 num_tiles = tilemap->num_tiles;
	UINT32 cached_indx;
	UINT8 *visible = tilemap->visible;
	UINT8 *dirty_vram = tilemap->dirty_vram;
	UINT8 *dirty_pixels = tilemap->dirty_pixels;

	memset( &cave_tile_info, 0x00, sizeof(cave_tile_info) ); /* initialize defaults */

	for( cached_indx=0; cached_indx<num_tiles; cached_indx++ )
	{
		if( visible[cached_indx] && dirty_vram[cached_indx] )
		{
			struct cached_tile_info *cached_tile_info = &tilemap->cached_tile_info[cached_indx];
			UINT32 memory_offset = tilemap->cached_indx_to_memory_offset[cached_indx];
			tilemap->tile_get_info( memory_offset );
			{
				UINT32 flags = cave_tile_info.flags;
				cached_tile_info->flags = (flags&0xfc)|logical_flip_to_cached_flip[flags&0x3];
			}
			cached_tile_info->pen_usage = cave_tile_info.pen_usage;
			cached_tile_info->pen_data = cave_tile_info.pen_data;
			cached_tile_info->pal_data = cave_tile_info.pal_data;
			tilemap->priority[cached_indx] = cave_tile_info.priority;
			dirty_pixels[cached_indx] = 1;
			dirty_vram[cached_indx] = 0;
			render_mask( tilemap, cached_indx );
		}
	}
}

static void update_visible( struct cave_tilemap *tilemap )
{
	// visibility marking is not currently implemented
	memset( tilemap->visible, 1, tilemap->num_tiles );
}

static void cave_tilemap_update( struct cave_tilemap *tilemap )
{
profiler_mark(PROFILER_TILEMAP_UPDATE);
	if( tilemap==CAVE_ALL_TILEMAPS )
	{
		tilemap = first_tilemap;
		while( tilemap )
		{
			cave_tilemap_update( tilemap );
			tilemap = tilemap->next;
		}
	}
	else if( tilemap->enable )
	{
		tilemap->bNeedRender = 1;
		update_visible( tilemap );
		update_tile_info( tilemap );
	}
profiler_mark(PROFILER_END);
}

/***********************************************************************************/

static void cave_tilemap_set_scrolldx( struct cave_tilemap *tilemap, int dx, int dx_if_flipped )
{
	tilemap->dx = dx;
	tilemap->dx_if_flipped = dx_if_flipped;
	tilemap->scrollx_delta = ( tilemap->attributes & CAVE_TILEMAP_FLIPX )?dx_if_flipped:dx;
}

static void cave_tilemap_set_scrolldy( struct cave_tilemap *tilemap, int dy, int dy_if_flipped )
{
	tilemap->dy = dy;
	tilemap->dy_if_flipped = dy_if_flipped;
	tilemap->scrolly_delta = ( tilemap->attributes & CAVE_TILEMAP_FLIPY )?dy_if_flipped:dy;
}

static void cave_tilemap_set_scrollx( struct cave_tilemap *tilemap, int which, int value )
{
	value = tilemap->scrollx_delta-value;

	if( tilemap->orientation & ORIENTATION_SWAP_XY )
	{
		if( tilemap->orientation & ORIENTATION_FLIP_X ) which = tilemap->scroll_cols-1 - which;
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) value = screen_height-tilemap->cached_height-value;
		if( tilemap->colscroll[which]!=value )
		{
			tilemap->colscroll[which] = value;
		}
	}
	else
	{
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) which = tilemap->scroll_rows-1 - which;
		if( tilemap->orientation & ORIENTATION_FLIP_X ) value = screen_width-tilemap->cached_width-value;
		if( tilemap->rowscroll[which]!=value )
		{
			tilemap->rowscroll[which] = value;
		}
	}
}

static void cave_tilemap_set_scrollx_dy( struct cave_tilemap *tilemap, int which, int value, int valuedy )
{
	value = tilemap->scrollx_delta-value;

	if( tilemap->orientation & ORIENTATION_SWAP_XY ){
		if( tilemap->orientation & ORIENTATION_FLIP_X ){
			which = tilemap->scroll_cols-1 - which;
			valuedy = -valuedy;
		}
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) value = screen_height-tilemap->cached_height-value;
		tilemap->colscroll[which] = (value&0xffff)|(valuedy<<16);
	}
	else {
		if( tilemap->orientation & ORIENTATION_FLIP_Y ){
			which = tilemap->scroll_rows-1 - which;
			valuedy = -valuedy;
		}
		if( tilemap->orientation & ORIENTATION_FLIP_X ) value = screen_width-tilemap->cached_width-value;
		tilemap->rowscroll[which] = (value&0xffff)|(valuedy<<16);
	}
}

static void cave_tilemap_set_scrolly( struct cave_tilemap *tilemap, int which, int value )
{
	value = tilemap->scrolly_delta - value;

	if( tilemap->orientation & ORIENTATION_SWAP_XY )
	{
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) which = tilemap->scroll_rows-1 - which;
		if( tilemap->orientation & ORIENTATION_FLIP_X ) value = screen_width-tilemap->cached_width-value;
		if( tilemap->rowscroll[which]!=value )
		{
			tilemap->rowscroll[which] = value;
		}
	}
	else
	{
		if( tilemap->orientation & ORIENTATION_FLIP_X ) which = tilemap->scroll_cols-1 - which;
		if( tilemap->orientation & ORIENTATION_FLIP_Y ) value = screen_height-tilemap->cached_height-value;
		if( tilemap->colscroll[which]!=value )
		{
			tilemap->colscroll[which] = value;
		}
	}
}
/***********************************************************************************/

static void cave_tilemap_draw( struct osd_bitmap *dest, struct cave_tilemap *tilemap, UINT32 flags )
{
	int xpos,ypos;
profiler_mark(PROFILER_TILEMAP_DRAW);
	tmap_render( tilemap );

	if( tilemap->enable )
	{
		void (*draw)( int, int );

		int rows = tilemap->scroll_rows;
		const int *rowscroll = tilemap->rowscroll;
		int cols = tilemap->scroll_cols;
		const int *colscroll = tilemap->colscroll;

		int left = tilemap->clip_left;
		int right = tilemap->clip_right;
		int top = tilemap->clip_top;
		int bottom = tilemap->clip_bottom;

		int tile_size = tilemap->tile_size;

		tmblt.screen = dest;
		if( dest )
		{
			tmblt.pixmap = tilemap->pixmap;
			tmblt.source_line_offset = tilemap->pixmap_line_offset;

			tmblt.dest_line_offset = dest->line[1] - dest->line[0];

			tmblt.dest_line_offset /= 2;
			tmblt.source_line_offset /= 2;

			tmblt.dest_row_offset = tile_size*tmblt.dest_line_offset;
		}
		else return;

		{
			draw = tilemap->draw;
			tmblt.bitmask = tilemap->foreground->bitmask;
			tmblt.mask_line_offset = tilemap->foreground->line_offset;
			tmblt.mask_data_row = tilemap->foreground->data_row;
			tmblt.mask_row_offset = tile_size*tmblt.mask_line_offset;
		}

		tmblt.source_row_offset = tile_size*tmblt.source_line_offset;

		tmblt.priority_data_row = tilemap->priority_row;
		tmblt.source_width = tilemap->cached_width;
		tmblt.source_height = tilemap->cached_height;
		tmblt.tile_priority = flags&0xf;

		if( rows == 1 && cols == 1 )
		{ /* XY scrolling playfield */
			int scrollx = rowscroll[0];
			int scrolly = colscroll[0];

			if( scrollx < 0 )
			{
				scrollx = tmblt.source_width - (-scrollx) % tmblt.source_width;
			}
			else
			{
				scrollx = scrollx % tmblt.source_width;
			}

			if( scrolly < 0 )
			{
				scrolly = tmblt.source_height - (-scrolly) % tmblt.source_height;
			}
			else
			{
				scrolly = scrolly % tmblt.source_height;
			}

	 		tmblt.clip_left = left;
	 		tmblt.clip_top = top;
	 		tmblt.clip_right = right;
	 		tmblt.clip_bottom = bottom;

			for(
				ypos = scrolly - tmblt.source_height;
				ypos < tmblt.clip_bottom;
				ypos += tmblt.source_height )
			{
				for(
					xpos = scrollx - tmblt.source_width;
					xpos < tmblt.clip_right;
					xpos += tmblt.source_width )
				{
					draw( xpos,ypos );
				}
			}
		}
		else if( rows == 1 )
		{ /* scrolling columns + horizontal scroll */
			int col = 0;
			int colwidth = (right-left) / cols;

			tmblt.clip_top = top;
			tmblt.clip_bottom = bottom;

			while( col < cols )
			{
				int cons = 1;
				int scrolly = colscroll[col];
				int scrollx = rowscroll[0] + (scrolly>>16);

				if( scrollx < 0 )
				{
					scrollx = tmblt.source_width - (-scrollx) % tmblt.source_width;
				}
				else
				{
					scrollx = scrollx % tmblt.source_width;
				}

	 			/* count consecutive columns scrolled by the same amount */
				{
					while( col + cons < cols &&	colscroll[col + cons] == scrolly ) cons++;

					if( scrolly < 0 )
					{
						scrolly = tmblt.source_height - (-scrolly) % tmblt.source_height;
					}
					else
					{
						scrolly %= tmblt.source_height;
					}

					tmblt.clip_left = left + col * colwidth;
					tmblt.clip_right = tmblt.clip_left + cons * colwidth;
					if (tmblt.clip_right > right) tmblt.clip_right = right;

					for(
						ypos = scrolly - tmblt.source_height;
						ypos < tmblt.clip_bottom;
						ypos += tmblt.source_height )
					{
						draw( scrollx,ypos );
					}

					for(
						ypos = scrolly - tmblt.source_height;
						ypos < tmblt.clip_bottom;
						ypos += tmblt.source_height )
					{
						draw( scrollx - tmblt.source_width,ypos );
					}
				}
				col += cons;
			}
		}
		else if( cols == 1 )
		{ /* scrolling rows + vertical scroll */
			int row = 0;
			int rowheight = (bottom - top) / rows;
			tmblt.clip_left = left;
			tmblt.clip_right = right;
			while( row < rows )
			{
				int cons = 1;
				int scrollx = rowscroll[row];
				int scrolly = colscroll[0] + (scrollx>>16);
				if( scrolly < 0 )
				{
					scrolly = tmblt.source_height - (-scrolly) % tmblt.source_height;
				}
				else
				{
					scrolly = scrolly % tmblt.source_height;
				}
				/* count consecutive rows scrolled by the same amount */
				{
					while( row + cons < rows &&	rowscroll[row + cons] == scrollx ) cons++;
					if( scrollx < 0)
					{
						scrollx = tmblt.source_width - (-scrollx) % tmblt.source_width;
					}
					else
					{
						scrollx %= tmblt.source_width;
					}
					tmblt.clip_top = top + row * rowheight;
					tmblt.clip_bottom = tmblt.clip_top + cons * rowheight;
					if (tmblt.clip_bottom > bottom) tmblt.clip_bottom = bottom;
					for(
						xpos = scrollx - tmblt.source_width;
						xpos < tmblt.clip_right;
						xpos += tmblt.source_width )
					{
						draw( xpos,scrolly );
					}
					for(
						xpos = scrollx - tmblt.source_width;
						xpos < tmblt.clip_right;
						xpos += tmblt.source_width )
					{
						draw( xpos,scrolly - tmblt.source_height );
					}
				}
				row += cons;
			}
		}
	}
profiler_mark(PROFILER_END);
}

/***********************************************************************************/

#else // DECLARE
DECLARE( draw, (int xpos, int ypos),
{
	struct osd_bitmap *screen = tmblt.screen;
	int x1 = xpos;
	int y1 = ypos;
	int x2 = xpos+tmblt.source_width;
	int y2 = ypos+tmblt.source_height;
	DATA_TYPE *dest_baseaddr = NULL;
	DATA_TYPE *dest_next;
	int dy;
	const DATA_TYPE *source0;
	DATA_TYPE *dest0;
	int i;
	int row;
	UINT8 *priority_data;
	int tile_type;
	int prev_tile_type;
	int x_start;
	int x_end;
	int column;
	int c1;
	int c2; /* leftmost and rightmost visible columns in source tilemap */
	int y; /* current screen line to render */
	int y_next;
	int num_pixels;
	int priority;
	const DATA_TYPE *source_baseaddr;
	const DATA_TYPE *source_next;
	const UINT8 *mask0;
	UINT8 *mask_data;
	const UINT8 *mask_baseaddr;
	const UINT8 *mask_next;

	/* clip source coordinates */
	if( x1<tmblt.clip_left ) x1 = tmblt.clip_left;
	if( x2>tmblt.clip_right ) x2 = tmblt.clip_right;
	if( y1<tmblt.clip_top ) y1 = tmblt.clip_top;
	if( y2>tmblt.clip_bottom ) y2 = tmblt.clip_bottom;

	if( x1<x2 && y1<y2 )
	{ /* do nothing if totally clipped */
		priority = tmblt.tile_priority;
		dest_baseaddr = xpos + (DATA_TYPE *)screen->line[y1];

		/* convert screen coordinates to source tilemap coordinates */
		x1 -= xpos;
		y1 -= ypos;
		x2 -= xpos;
		y2 -= ypos;

		source_baseaddr = (DATA_TYPE *)tmblt.pixmap->line[y1];
		mask_baseaddr = tmblt.bitmask->line[y1];

		c1 = x1/TILE_SIZE; /* round down */
		c2 = (x2+TILE_SIZE-1)/TILE_SIZE; /* round up */

		y = y1;
		y_next = TILE_SIZE*(y1/TILE_SIZE) + TILE_SIZE;
		if( y_next>y2 ) y_next = y2;

		dy = y_next-y;
		dest_next = dest_baseaddr + dy*tmblt.dest_line_offset;
		source_next = source_baseaddr + dy*tmblt.source_line_offset;
		mask_next = mask_baseaddr + dy*tmblt.mask_line_offset;
		for(;;)
		{
			row = y/TILE_SIZE;
			mask_data = tmblt.mask_data_row[row];
			priority_data = tmblt.priority_data_row[row];
			prev_tile_type = TILE_TRANSPARENT;
			x_start = x1;

			for( column=c1; column<=c2; column++ )
			{
				if( column==c2 || priority_data[column]!=priority )
				{
					tile_type = TILE_TRANSPARENT;
				}
				else
				{
					tile_type = mask_data[column];
				}

				if( tile_type!=prev_tile_type )
				{
					x_end = column*TILE_SIZE;
					if( x_end<x1 ) x_end = x1;
					if( x_end>x2 ) x_end = x2;

					if( prev_tile_type != TILE_TRANSPARENT )
					{
						if( prev_tile_type == TILE_MASKED )
						{
							const UINT8 *mask1;
							const DATA_TYPE *source1;
							DATA_TYPE *dest1;
							int j;
							int cnt = x_end-x_start;
							int cnt1 = (-x_start)&7;
							int cnt2;
							int cnt3;

							i = y;
							mask0 = mask_baseaddr;
							source0 = source_baseaddr;
							dest0 = dest_baseaddr;

							if(cnt1>=cnt)
							{
								cnt1=cnt;cnt2=0;cnt3=0;
							}
							else
							{
								cnt3 = x_end&7;
								if(cnt1+cnt3==cnt)
								{
									cnt2=0;
								}
								else
								{
									cnt2=(cnt-cnt1)/8;
								}
							}
							for(;;)
							{
								int xp=x_start;

								for(j=cnt1;j>0;j--)
								{
									if( mask0[xp/8]&(0x80>>(xp&7)) ) dest0[xp] = source0[xp];
									xp++;
								}

								if(cnt2)
								{
									source1 = &source0[xp];
									dest1 = &dest0[xp];
									mask1 = &mask0[xp/8];
									for(j=cnt2;j>0;j--)
									{
										UINT8 data = *mask1++;
										if( data&0x80 ) dest1[0] = source1[0];
										if( data&0x40 ) dest1[1] = source1[1];
										if( data&0x20 ) dest1[2] = source1[2];
										if( data&0x10 ) dest1[3] = source1[3];
										if( data&0x08 ) dest1[4] = source1[4];
										if( data&0x04 ) dest1[5] = source1[5];
										if( data&0x02 ) dest1[6] = source1[6];
										if( data&0x01 ) dest1[7] = source1[7];
										source1+=8;
										dest1+=8;
									}
									xp+=cnt2*8;
								}

								for(j=cnt3;j>0;j--)
								{
									if( mask0[xp/8]&(0x80>>(xp&7)) ) dest0[xp] = source0[xp];
									xp++;
								}

								if( ++i == y_next ) break;
								dest0 += tmblt.dest_line_offset;
								source0 += tmblt.source_line_offset;
								mask0 += tmblt.mask_line_offset;
							}
						}
						else
						{ /* TILE_OPAQUE */
							num_pixels = x_end - x_start;
							dest0 = dest_baseaddr+x_start;
							source0 = source_baseaddr+x_start;
							i = y;
							for(;;)
							{
								memcpy( dest0, source0, num_pixels*sizeof(DATA_TYPE) );
								if( ++i == y_next ) break;

								dest0 += tmblt.dest_line_offset;
								source0 += tmblt.source_line_offset;
							}
						}
					}
					x_start = x_end;
				}

				prev_tile_type = tile_type;
			}

			if( y_next==y2 ) break; /* we are done! */

			dest_baseaddr = dest_next;
			source_baseaddr = source_next;
			mask_baseaddr = mask_next;
			y = y_next;
			y_next += TILE_SIZE;

			if( y_next>=y2 )
			{
				y_next = y2;
			}
			else
			{
				dest_next += tmblt.dest_row_offset;
				source_next += tmblt.source_row_offset;
				mask_next += tmblt.mask_row_offset;
			}
		} /* process next row */
	} /* not totally clipped */
})



#if (TILE_SIZE == 8) /* only construct once for each depth */
DECLARE( draw_tile, (struct cave_tilemap *tilemap, UINT32 cached_indx, UINT32 col, UINT32 row ),
{
	struct cached_tile_info *cached_tile_info = &tilemap->cached_tile_info[cached_indx];
	UINT32 tile_size = tilemap->tile_size;
	const UINT8 *pPenData = cached_tile_info->pen_data;
	int pitch = tile_size;
	const UINT32 *pPalData = cached_tile_info->pal_data;
	UINT32 flags = cached_tile_info->flags;
	UINT32 y0 = tile_size*row;
	UINT32 x0 = tile_size*col;
	struct osd_bitmap *pPixmap = tilemap->pixmap;
	UINT32 *pPenToPixel = tilemap->pPenToPixel[flags&(CAVE_TILE_SWAPXY|CAVE_TILE_FLIPY|CAVE_TILE_FLIPX)];
	int tx;
	int ty;
	const UINT8 *pSource;
	UINT8 data;
	UINT32 yx;

	if( flags&CAVE_TILE_4BPP )
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size/2; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				*(x0+(yx%MAX_TILESIZE)+(DATA_TYPE *)pPixmap->line[y0+yx/MAX_TILESIZE]) = pPalData[data&0xf];
				yx = *pPenToPixel++;
				*(x0+(yx%MAX_TILESIZE)+(DATA_TYPE *)pPixmap->line[y0+yx/MAX_TILESIZE]) = pPalData[data>>4];
			}
			pPenData += pitch/2;
		}
	}
	else
	{
		for( ty=tile_size; ty!=0; ty-- )
		{
			pSource = pPenData;
			for( tx=tile_size; tx!=0; tx-- )
			{
				data = *pSource++;
				yx = *pPenToPixel++;
				*(x0+(yx%MAX_TILESIZE)+(DATA_TYPE *)pPixmap->line[y0+yx/MAX_TILESIZE]) = pPalData[data];
			}
			pPenData += pitch;
		}
	}
})

#endif /* draw_tile */

#undef TILE_SIZE
#undef DATA_TYPE
#undef DECLARE

#endif /* DECLARE */
