/***************************************************************************

  vidhrdw.c

  Functions to emulate the video hardware of the machine.

Important!	There are two types of NeoGeo romdump - MVS & MGD2.  They are both
converted to a standard format in the vh_start routines.


Graphics information:

0x00000 - 0xdfff	: Blocks of sprite data, each 0x80 bytes:
	Each 0x80 block is made up of 0x20 double words, their format is:
	Word: Sprite number (16 bits)
	Byte: Palette number (8 bits)
	Byte: Bit 0: X flip
		  Bit 1: Y flip
		  Bit 2: Automatic animation flag (4 tiles?)
		  Bit 3: Automatic animation flag (8 tiles?)
		  Bit 4: MSB of sprite number (confirmed, Karnov_r, Mslug). See note.
		  Bit 5: MSB of sprite number (MSlug2)
		  Bit 6: MSB of sprite number (Kof97)
		  Bit 7: Unknown for now

	Each double word sprite is drawn directly underneath the previous one,
	based on the starting coordinates.

0x0e000 - 0x0ea00	: Front plane fix tiles (8*8), 2 bytes each

0x10000: Control for sprites banks, arranged in words

	Bit 0 to 3 - Y zoom LSB
	Bit 4 to 7 - Y zoom MSB (ie, 1 byte for Y zoom).
	Bit 8 to 11 - X zoom, 0xf is full size (no scale).
	Bit 12 to 15 - Unknown, probably unused

0x10400: Control for sprite banks, arranged in words

	Bit 0 to 5: Number of sprites in this bank (see note below).
	Bit 6 - If set, this bank is placed to right of previous bank (same Y-coord).
	Bit 7 to 15 - Y position for sprite bank.

0x10800: Control for sprite banks, arranged in words
	Bit 0 to 5: Unknown
	Bit 7 to 15 - X position for sprite bank.

Notes:

* If rom set has less than 0x10000 tiles then msb of tile must be ignored
(see Magician Lord).

***************************************************************************/

#include "driver.h"
#include "common.h"
#include "usrintrf.h"
#include "vidhrdw/generic.h"


static data16_t *neogeo_vidram16;
static data16_t *neogeo_paletteram16;	/* pointer to 1 of the 2 palette banks */
static data16_t *neogeo_palettebank[2]; /* 0x100*16 2 byte palette entries */
static int neogeo_palette_index;
static data16_t neogeo_vidram16_modulo;
static data16_t neogeo_vidram16_offset;
static int high_tile;
static int vhigh_tile;
static int vvhigh_tile;
int no_of_tiles;
static int palette_swap_pending;
static int fix_bank;

extern data16_t *neogeo_ram16;
extern unsigned int neogeo_frame_counter;
extern int neogeo_game_fix;

UINT32 *l_y_skip;
static UINT32 dda_y_skip[0x100][0x21][17];
static UINT8 yskip[0x100][0x21];
static UINT8 mmy[0x100][0x10];

#ifdef UOMAME32J
extern int neogeo_machine_country;
int gururin_fix = 0;
int sdodgeb_fix = 0;
#endif

void neogeo_init_sprite(void);
void (*neo_draw_sprite[2][16])(UINT32 *sprite, UINT16 *bitmap, UINT32 *paldata, int zy, int dy);

/******************************************************************************/

static void build_zoom_table(void)
{
	int i, my, zy, dday;

	for (zy = 0; zy <= 0xff; zy++)
	{
		dday = 0;

		for (my = 0; my < 0x21; my++)
		{
			if (zy != 0xff)
			{
				yskip[zy][my] = 0;
				dda_y_skip[zy][my][0] = 0;

				for(i = 0; i < 16; i++)
				{
					dda_y_skip[zy][my][i + 1] = 0;
					dday -= zy + 1;
					if (dday <= 0)
					{
						dday += 256;
						yskip[zy][my]++;
					}
					dda_y_skip[zy][my][yskip[zy][my]]++;
				}
			}
			else
			{
				dda_y_skip[zy][my][0] = 0;

				for(i = 1; i < 16; i++)
					dda_y_skip[zy][my][i] = 1;

				yskip[zy][my] = 16;
			}

			if (my < 0x10)
			{
				mmy[zy][my] = ((my * 16 * 256) / (zy + 1) + 15) / 16;
				if (mmy[zy][my] > 0x10)
					mmy[zy][my] = 0x10;
			}
		}
	}
}

/******************************************************************************/

void neogeo_vh_stop(void)
{
	if (neogeo_palettebank[0])
		free (neogeo_palettebank[0]);
	neogeo_palettebank[0] = NULL;

	if (neogeo_palettebank[1])
		free (neogeo_palettebank[1]);
	neogeo_palettebank[1] = NULL;

	if (neogeo_vidram16)
		free (neogeo_vidram16);
	neogeo_vidram16 = NULL;

	if (neogeo_ram16)
		free (neogeo_ram16);
	neogeo_ram16 = NULL;
}

static int common_vh_start(void)
{
	neogeo_palettebank[0] = NULL;
	neogeo_palettebank[1] = NULL;
	neogeo_vidram16 = NULL;

	neogeo_palettebank[0] = malloc(0x2000);
	if (!neogeo_palettebank[0])
	{
		neogeo_vh_stop();
		return 1;
	}

	neogeo_palettebank[1] = malloc(0x2000);
	if (!neogeo_palettebank[1])
	{
		neogeo_vh_stop();
		return 1;
	}

	/* 0x20000 bytes even though only 0x10c00 is used */
	neogeo_vidram16 = malloc(0x20000);
	if (!neogeo_vidram16)
	{
		neogeo_vh_stop();
		return 1;
	}
	memset(neogeo_vidram16,0,0x20000);

	neogeo_paletteram16 = neogeo_palettebank[0];
	neogeo_palette_index = 0;
	neogeo_vidram16_modulo = 1;
	neogeo_vidram16_offset = 0;
	fix_bank = 0;
	palette_swap_pending = 0;

	build_zoom_table();

	neogeo_init_sprite();

	return 0;
}


static void decodetile(int tileno)
{
	unsigned char swap[128];
	UINT32 *gfxdata;
	int x,y;
	unsigned int pen;


	gfxdata = (UINT32 *)&memory_region(REGION_GFX2)[128 * tileno];

	memcpy(swap,gfxdata,128);

	for (y = 0; y < 16; y++)
	{
		UINT32 dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[64 + 4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[64 + 4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[64 + 4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[64 + 4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;

		dw = 0;
		for (x = 0; x < 8; x++)
		{
			pen  = ((swap[4*y + 3] >> x) & 1) << 3;
			pen |= ((swap[4*y + 1] >> x) & 1) << 2;
			pen |= ((swap[4*y + 2] >> x) & 1) << 1;
			pen |=  (swap[4*y    ] >> x) & 1;
			dw |= pen << 4*x;
			Machine->gfx[2]->pen_usage[tileno] |= (1 << pen);
		}
		*(gfxdata++) = dw;
	}
}

int neogeo_mvs_vh_start(void)
{
	no_of_tiles=memory_region_length(REGION_GFX2)/128;
	if (no_of_tiles>0x10000) high_tile=1; else high_tile=0;
	if (no_of_tiles>0x20000) vhigh_tile=1; else vhigh_tile=0;
	if (no_of_tiles>0x40000) vvhigh_tile=1; else vvhigh_tile=0;
	Machine->gfx[2]->total_elements = no_of_tiles;
	if (Machine->gfx[2]->pen_usage)
		free(Machine->gfx[2]->pen_usage);
	Machine->gfx[2]->pen_usage = malloc(no_of_tiles * sizeof(int));
	memset(Machine->gfx[2]->pen_usage,0,no_of_tiles * sizeof(int));

	/* tiles are not decoded yet. They will be decoded later as they are used. */
	/* pen_usage is used as a marker of decoded tiles: if it is 0, then the tile */
	/* hasn't been decoded yet. */

	return common_vh_start();
}

/******************************************************************************/

static void swap_palettes(void)
{
	int i;

	for (i = 0; i < 0x2000 >> 1; i++)
	{
		data16_t newword = neogeo_paletteram16[i];
		int r,g,b;

		r = ((newword >> 7) & 0x1e) | ((newword >> 14) & 0x01);
		g = ((newword >> 3) & 0x1e) | ((newword >> 13) & 0x01);
		b = ((newword << 1) & 0x1e) | ((newword >> 12) & 0x01);

		r = (r << 3) | (r >> 2);
		g = (g << 3) | (g >> 2);
		b = (b << 3) | (b >> 2);

		palette_change_color(i, r, g, b);
	}

	palette_swap_pending = 0;
}

static void neogeo_setpalbank(int n)
{
	if (neogeo_palette_index != n)
	{
		neogeo_palette_index = n;
		neogeo_paletteram16 = neogeo_palettebank[n];
		palette_swap_pending = 1;
	}
}

WRITE16_HANDLER( neogeo_setpalbank0_16_w )
{
	neogeo_setpalbank(0);
}

WRITE16_HANDLER( neogeo_setpalbank1_16_w )
{
	neogeo_setpalbank(1);
}

READ16_HANDLER( neogeo_paletteram16_r )
{
	return neogeo_paletteram16[offset];
}

WRITE16_HANDLER( neogeo_paletteram16_w )
{
	data16_t oldword, newword;
	int r,g,b;

	oldword = newword = neogeo_paletteram16[offset];
	COMBINE_DATA(&newword);

	if (oldword == newword)
		return;

	neogeo_paletteram16[offset] = newword;

	r = ((newword >> 7) & 0x1e) | ((newword >> 14) & 0x01);
	g = ((newword >> 3) & 0x1e) | ((newword >> 13) & 0x01);
	b = ((newword << 1) & 0x1e) | ((newword >> 12) & 0x01);

	r = (r << 3) | (r >> 2);
	g = (g << 3) | (g >> 2);
	b = (b << 3) | (b >> 2);

	palette_change_color(offset, r, g, b);
}

/******************************************************************************/

WRITE16_HANDLER( neogeo_vidram16_offset_w )
{
	COMBINE_DATA(&neogeo_vidram16_offset);
}

READ16_HANDLER( neogeo_vidram16_data_r )
{
	return neogeo_vidram16[neogeo_vidram16_offset];
}

WRITE16_HANDLER( neogeo_vidram16_data_w )
{
	COMBINE_DATA(&neogeo_vidram16[neogeo_vidram16_offset]);
	neogeo_vidram16_offset += neogeo_vidram16_modulo;
	neogeo_vidram16_offset &= 0xffff;
}

/* Modulo can become negative , Puzzle Bobble Super Sidekicks and a lot */
/* of other games use this */
WRITE16_HANDLER( neogeo_vidram16_modulo_w )
{
	COMBINE_DATA(&neogeo_vidram16_modulo);
}

READ16_HANDLER( neogeo_vidram16_modulo_r )
{
	return neogeo_vidram16_modulo;
}


WRITE16_HANDLER( neo_board_fix_16_w )
{
	fix_bank = 1;
}

WRITE16_HANDLER( neo_game_fix_16_w )
{
	fix_bank = 0;
}

/******************************************************************************/

INLINE void neo_draw_gfx(unsigned char **line, UINT32 *paldata,
		unsigned int code, int flipx, int flipy, int sx, int sy, int zx, int zy,
		const struct rectangle *clip)
{
	int oy, ey, dy;
	UINT32 *fspr;

	oy = sy;
	ey = sy + zy -1;

	if (sy < clip->min_y) sy = clip->min_y;
	if (ey >= clip->max_y) ey = clip->max_y;

	fspr = (UINT32 *)memory_region(REGION_GFX2);

	if (!flipy) /* normal */
	{
		dy = 2;
		fspr += code * 32 + (sy - oy) * 2;
	}
	else /* Y flip */
	{
		dy = -2;
		fspr += (code + 1) * 32 - 2 - (sy - oy) * 2;
	}

	zy = ey - sy + 1;

	(*neo_draw_sprite[flipx][zx - 1])(fspr, (UINT16 *)&line[sy][sx*2], paldata, zy, dy);
}


void NeoMVSDrawGfx(unsigned char **line,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		int zx,int zy,const struct rectangle *clip)
{
	/* Safety feature */
	code = code % no_of_tiles;

	if (!gfx->pen_usage[code])	/* decode tile if it hasn't been yet */
		decodetile(code);

	/* Check for total transparency, no need to draw */
	if (gfx->pen_usage[code] & ~1)
	{
		neo_draw_gfx(line,
			&gfx->colortable[gfx->color_granularity * color],
			code,
			flipx, flipy,
			sx,sy,zx,zy,
			clip
		);
	}
}



#ifdef UOMAME32J
INLINE int convert_jfont(int code)
{
	if (sdodgeb_fix)
	{
		switch (code)
		{
		case 0x016e:
		case 0x0166:
		case 0x0173:
			code -= 0xc0;
			break;

		case 0x026e:
		case 0x0273:
		case 0x0266:
			code -= 0x180;
			break;

		case 0x010d:
			code = 0x008d;
			break;
		case 0x020d:
			code = 0x00cd;
			break;
		}
	}

	if (code >= 0x100 && code <= 0x2ff)
		code += 0x200;

	return code;
}
#endif


/******************************************************************************/

static void screenrefresh(struct osd_bitmap *bitmap, const struct rectangle *clip)
{
	int sx = 0, sy = 0, oy = 0, my = 0, zx = 0, zy = 0;
	int offs, count, y, x;
	int tileno, tileatr, t1, t3;
	char fullmode = 0;
	unsigned char **line = bitmap->line;
	unsigned int *pen_usage;
	struct GfxElement *gfx = Machine->gfx[2]; /* Save constant struct dereference */

	/* Palette swap occured after last frame but before this one */
	if (palette_swap_pending) swap_palettes();

	/* Do compressed palette stuff */
	palette_recalc();
	/* no need to check the return code since we redraw everything each frame */

	fillbitmap(bitmap,Machine->pens[4095],clip);

	/* Draw sprites */
	for (count = 0; count < 0x300 >> 1; count++)
	{
		t3 = neogeo_vidram16[(0x10000 >> 1) + count];
		t1 = neogeo_vidram16[(0x10400 >> 1) + count];

		/* If this bit is set this new column is placed next to last one */
		if (t1 & 0x40)
		{
			/* No point doing anything if tile strip is 0 */
			if (my == 0) continue;

			sx += zx;
			/* Get new zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
				zx = 0x0f;
			zx++;
			sy = oy;
		}
		else /* nope it is a new block */
		{
			/* Number of tiles in this strip */
			my = t1 & 0x3f;
			if (my == 0)
				continue; /* No point doing anything if tile strip is 0 */
			else if (my == 0x20)
				fullmode = 1;
			else if (my >= 0x21)
				fullmode = 2;	/* most games use 0x21, but  Alpha Mission II uses 0x3f */
			else
				fullmode = 0;

			/* Sprite scaling */
			zy = t3 & 0xff;
			/* Get new X zoom for this column */
			zx = (t3 >> 8) & 0x0f;
			if (gururin_fix && (t3 == 0 || t3 == 0x147f))
			{
				zx = 0x0f;
				zy = 0xff;
			}
			if (zy == 0)
			{
				/* No point doing anything if tile Y zoom is 0 */
				my = 0;
				continue;
			}
			zx++;

			sy = 0x200 - (t1 >> 7);

			if (clip->max_y - clip->min_y > 8 ||	/* kludge to improve the ssideki games */
				clip->min_y == Machine->visible_area.min_y)
			{
				if (sy > 0x110) sy -= 0x200;
				if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
				{
					while (sy < 0) sy += 2 * (zy + 1);
				}
			}
			oy = sy;

			if (zy < 0xff && my < 0x10)
				my = mmy[zy][my];
			else if (my > 0x20)
				my = 0x20;

			sx = (neogeo_vidram16[(0x10800 >> 1) + count] >> 7);
		}

		if (sx >= 320)
			sx -= 0x200;

		if (sx + zx <= 8)
			continue;

		offs = count << 6;

		/* my holds the number of tiles in each vertical multisprite block */
		for (y = 0; y < my ; y++)
		{
			if (fullmode == 2 || (fullmode == 1 && zy == 0xff))
			{
				if (sy >= 248) sy -= 2 * (zy + 1);
			}
			else if (fullmode == 1)
			{
				if (y == 0x10) sy -= 2 * (zy + 1);
			}
			else if (sy > 0x110)
			{
				sy -= 0x200;	/* NS990105 mslug2 fix */
			}

			if (sy + yskip[zy][y] > clip->min_y && sy <= clip->max_y)
			{
				tileno  = neogeo_vidram16[offs++];
				tileatr = neogeo_vidram16[offs++];

				if (high_tile && tileatr & 0x10)   tileno += 0x10000;
				if (vhigh_tile && tileatr & 0x20)  tileno += 0x20000;
				if (vvhigh_tile && tileatr & 0x40) tileno += 0x40000;

				if (tileatr & 0x08)
				{
					tileno = (tileno & ~7) + (neogeo_frame_counter & 7);
				}
				else if (tileatr & 0x04)
				{
					tileno = (tileno & ~3) + (neogeo_frame_counter & 3);
				}

				l_y_skip = dda_y_skip[zy][y];

				/* Safety feature */
				tileno = tileno % no_of_tiles;

				if (!gfx->pen_usage[tileno])	/* decode tile if it hasn't been yet */
					decodetile(tileno);

				/* Check for total transparency, no need to draw */
				if (gfx->pen_usage[tileno] & ~1)
				{
					UINT32 *paldata = &gfx->colortable[gfx->color_granularity * (tileatr >> 8)];

					neo_draw_gfx(line,
						paldata,
						tileno,
						tileatr & 0x01,tileatr & 0x02,
						sx,sy,zx,yskip[zy][y],
						clip
					);
				}
			}
			else offs += 2;

			sy += yskip[zy][y];
		}  /* for y */
	}  /* for count */


	/* Save some struct de-refs */
	gfx = Machine->gfx[fix_bank];
	pen_usage=gfx->pen_usage;

	/* Character foreground */
	for (y=clip->min_y / 8; y <= clip->max_y / 8; y++)
	{
		for (x = 0; x < 40; x++)
		{

			int byte1 = neogeo_vidram16[(0xe000 >> 1) + y + 32 * x];
			int byte2 = byte1 >> 12;
			byte1 = byte1 & 0xfff;

#ifdef UOMAME32J
			if (fix_bank && !neogeo_machine_country)
				byte1 = convert_jfont(byte1);
#endif

			if ((pen_usage[byte1] & ~1) == 0) continue;

			drawgfx(bitmap,gfx,
					byte1,
					byte2,
					0,0,
					x*8,y*8,
					clip,TRANSPARENCY_PEN,0);
		}
	}
}

void neogeo_vh_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
	screenrefresh(bitmap,&Machine->visible_area);
}

static int next_update_first_line;
extern int neogeo_raster_enable;

void neogeo_vh_raster_partial_refresh(struct osd_bitmap *bitmap,int current_line)
{
	struct rectangle clip;

	if (current_line < next_update_first_line)
		next_update_first_line = 0;

	clip.min_x = Machine->visible_area.min_x;
	clip.max_x = Machine->visible_area.max_x;
	clip.min_y = next_update_first_line;
	clip.max_y = current_line;
	if (clip.min_y < Machine->visible_area.min_y)
		clip.min_y = Machine->visible_area.min_y;
	if (clip.max_y > Machine->visible_area.max_y)
		clip.max_y = Machine->visible_area.max_y;

	if (clip.max_y >= clip.min_y)
	{
//logerror("refresh %d-%d\n",clip.min_y,clip.max_y);
		screenrefresh(bitmap,&clip);

		if (neogeo_raster_enable == 2)
		{
			int x;
			for (x = clip.min_x;x <= clip.max_x;x++)
			{
				if (x & 8)
					plot_pixel(bitmap,x,clip.max_y,Machine->uifont->colortable[1]);
			}
		}
	}

	next_update_first_line = current_line + 1;

}

void neogeo_vh_raster_screenrefresh(struct osd_bitmap *bitmap,int full_refresh)
{
}

